/*
 * calculation.c: Calculation class
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#include "math.h"
#include "defs.h"
#include "setup.h"
#include "calculation.h"

#define ROUND_INT(d) ( (int) ((d) + ((d) > 0 ? 0.5 : -0.5)) )

#define AnalysisDepth		SolarstormSetup.AnalysisDepth
#define EdgeWeighting		SolarstormSetup.EdgeWeighting
#define MinInputBrightness	SolarstormSetup.MinInputBrightness
#define MinOutputBrightness	SolarstormSetup.MinOutputBrightness
#define BrightnessCorrection	SolarstormSetup.BrightnessCorrection
#define HistogramAnalysisWidth 	SolarstormSetup.HistogramAnalysisWidth

/* Anzahl der Segmente */
static uint8_t left, top, right, bottom;

/* Breite & Hoehe des Bildes */
static uint16_t width, height;

/* Breite / Hoehe eines Segments */
static uint16_t len_left, len_right, len_top, len_bottom;

/* Tiefe des Auswertebereichs */
int len_lr, len_tb;

/* Zeiger auf die Werte des Auswertebereichs */
int *lr;
int *tb;

tHSVColor *hsv_pic;

/* Berechnungsfunktionen */
void calc_left(tRGBColor *col);
void calc_right(tRGBColor *col);
void calc_top(tRGBColor *col);
void calc_bottom(tRGBColor *col);

cCalc::cCalc(uint8_t l, uint8_t t, uint8_t r, uint8_t b, uint16_t w, uint16_t h)
{
  left = l;
  top = t;
  right = r;
  bottom = b;
  width = w;
  height = h;

  len_left   = height / left;
  len_right  = height / right;
  len_top    = width / top;
  len_bottom = width / bottom;

  CalcWeight();
}

cCalc::~cCalc()
{
  delete [] lr;
  delete [] tb;
}

void cCalc::CalcWeight(void)
{
  len_lr = ROUND_INT(width  * (AnalysisDepth / 100.0f));
  len_tb = ROUND_INT(height * (AnalysisDepth / 100.0f));

  lr = new int[len_lr];
  tb = new int[len_tb];

  // Gewichtung left & right berechnen
  for (int i=0; i<len_lr; i++)
  {
    float col_norm = (float)i / ((float)width - 1.0); // [0;Width] -> [0;1]
    lr[i] = (int)(255.0 * (float)pow((1.0 - col_norm), EdgeWeighting));
  }

  // Gewichtung top & bottom berechnen
  for (int i=0; i<len_tb; i++)
  {
    float row_norm = (float)i / ((float)height - 1.0); // [0;Height] -> [0;1]
    tb[i] = (int)(255.0 * (float)pow(1.0 - row_norm, EdgeWeighting));
  }
}

void cCalc::CalcChannels(tRGBColor *col, tHSVColor *HSV_Img)
{
  hsv_pic = HSV_Img;

  calc_left(&col[0]);
  calc_right(&col[left+top]);
  calc_top(&col[left]);
  calc_bottom(&col[left+top+right]);
}

void calc_left(tRGBColor *col)
{
  /***************************************************************************/
  /* Hue                                                                     */
  /***************************************************************************/
  // Hue-Histogramm ermitteln links
  long int* hue_hist = new long int[360 * left];  // hue histogram
  memset(hue_hist, 0, 360*left*sizeof(long int)); // clean histogram

  for (int channel=0; channel<left; channel++) {
    for (int y=channel*len_left; y<(channel+1)*len_left; y++) {
      for (int x=0; x<len_lr; x++)
      {
        tHSVColor hsv = hsv_pic[y*width + x];
        // perform calculations only if pixel has some luminosity (no black bars etc.)
        if (hsv.v > MinInputBrightness)
        {
          // Add weight to channel
          hue_hist[360*channel + hsv.h] += lr[x] * hsv.v;
        }
      }
    }
  }

  // Hue-Wert ermitteln links
  int* most_used_hue = new int[left];         // resulting hue for each channel
  memset(most_used_hue, 0, left*sizeof(int)); // clean

  for (int channel=0; channel<left; channel++)
  {
    int max = 0;
    for (int i = 0; i < 360; i++) // walk through histogram
    {
      int tmp = 0;
      for (int j = -HistogramAnalysisWidth; j < HistogramAnalysisWidth+1; j++)
      {
        int idx = i + j; // histogram item
        if (idx <   0) { idx = 360 - i; } // handle beginning of histogram -> roll back
        if (idx > 359) { idx = i; } // handle end of histogram -> roll forward

        tmp += hue_hist[360*channel + idx] * ((HistogramAnalysisWidth+1)-abs(j));
      }
      if (max < tmp)
      {
        max = tmp;
        most_used_hue[channel] = i;
      }
    }
  }

  /***************************************************************************/
  /* Saturation                                                              */
  /***************************************************************************/
  // Sat-Histogramm und Durchschnitts-Helligkeit (Value) ermitteln links
  long int* sat_hist = new long int[256 * left];  // sat histogram
  memset(sat_hist, 0, 256*left*sizeof(long int)); // clean histogram

  int* val_avg = new int[left];         // resulting value for each channel
  memset(val_avg, 0, left*sizeof(int)); // clean

  for (int channel=0; channel<left; channel++)
  {
    int val_cnt = 0;
    for (int y=channel*len_left; y<(channel+1)*len_left; y++)
    {
      for (int x=0; x<len_lr; x++)
      {
        tHSVColor hsv = hsv_pic[y*width + x];
        // perform calculations only if pixel has some luminosity (no black bars etc.)
        if (hsv.v > MinInputBrightness)
        {
          // find histogram position for pixel
          if ( (abs(most_used_hue[channel] - hsv.h) <= HistogramAnalysisWidth) ||
             (abs(most_used_hue[channel] - hsv.h) >= (360-HistogramAnalysisWidth)) )
          {
            // build histogram
            sat_hist[256*channel + hsv.s] += lr[x] * hsv.v;

            // build brightness average
            val_avg[channel] += hsv.v;
            val_cnt++;
          }
        }
      }
    }

    // calculate brightness average
    if (val_cnt > 0) { val_avg[channel] = val_avg[channel] / val_cnt; }
    else { val_avg[channel] = MinOutputBrightness; }
    // adjust brightness
    int new_value = (int) ((float)val_avg[channel] * ((float)BrightnessCorrection / 100.0));
    if (new_value > 255) { new_value = 255; } // ensure brightness isn't set too high
    val_avg[channel] = (unsigned char)new_value;
  }

  // Sat-Wert ermitteln links
  int* most_used_sat = new int[left];         // resulting sat for each channel
  memset(most_used_sat, 0, left*sizeof(int)); // clean

  for (int channel=0; channel<left; channel++)
  {
    int max = 0;
    for (int i = 0; i < 256; i++) // walk through histogram
    {
      int tmp = 0;
      for (int j = -HistogramAnalysisWidth; j < HistogramAnalysisWidth+1; j++)
      {
        int idx = i + j; // histogram item
        if (idx <   0) { idx = 256 - i; } // handle beginning of histogram -> roll back
        if (idx > 255) { idx = i; } // handle end of histogram -> roll forward

        tmp += sat_hist[256*channel + idx] * ((HistogramAnalysisWidth+1)-abs(j));
      }
      if (max < tmp)
      {
        max = tmp;
        most_used_sat[channel] = i;
      }
    }
  }

  /***************************************************************************/
  /* calculate results                                                       */
  /***************************************************************************/
  for (int channel = 0; channel < left; channel++) {
    col[left-channel-1] = HSV2RGB(most_used_hue[channel], most_used_sat[channel], val_avg[channel]); // convert back to rgb
  }

  delete [] most_used_hue;
  delete [] hue_hist;

  delete [] most_used_sat;
  delete [] sat_hist;

  delete [] val_avg;
}

void calc_right(tRGBColor *col)
{
  /***************************************************************************/
  /* Hue                                                                     */
  /***************************************************************************/
  // Hue-Histogramm ermitteln rechts
  long int* hue_hist = new long int[360 * right];  // hue histogram
  memset(hue_hist, 0, 360*right*sizeof(long int)); // clean histogram

  for (int channel=0; channel<right; channel++) {
    for (int y=channel*len_right; y<(channel+1)*len_right; y++) {
      for (int x=len_lr-1; x>=0; x--)
      {
        tHSVColor hsv = hsv_pic[y*width + width-x-1];
        // perform calculations only if pixel has some luminosity (no black bars etc.)
        if (hsv.v > MinInputBrightness)
        {
          // Add weight to channel
          hue_hist[360*channel + hsv.h] += lr[x] * hsv.v;
        }
      }
    }
  }

  // Hue-Wert ermitteln rechts
  int* most_used_hue = new int[right];         // resulting hue for each channel
  memset(most_used_hue, 0, right*sizeof(int)); // clean

  for (int channel=0; channel<right; channel++)
  {
    int max = 0;
    for (int i = 0; i < 360; i++) // walk through histogram
    {
      int tmp = 0;
      for (int j = -HistogramAnalysisWidth; j < HistogramAnalysisWidth+1; j++)
      {
        int idx = i + j; // histogram item
        if (idx <   0) { idx = 360 - i; } // handle beginning of histogram -> roll back
        if (idx > 359) { idx = i; } // handle end of histogram -> roll forward

        tmp += hue_hist[360*channel + idx] * ((HistogramAnalysisWidth+1)-abs(j));
      }
      if (max < tmp)
      {
        max = tmp;
        most_used_hue[channel] = i;
      }
    }
  }

  /***************************************************************************/
  /* Saturation                                                              */
  /***************************************************************************/
  // Sat-Histogramm und Durchschnitts-Helligkeit (Value) ermitteln rechts
  long int* sat_hist = new long int[256 * right];  // sat histogram
  memset(sat_hist, 0, 256*right*sizeof(long int)); // clean histogram

  int* val_avg = new int[right];         // resulting value for each channel
  memset(val_avg, 0, right*sizeof(int)); // clean

  for (int channel=0; channel<right; channel++)
  {
    int val_cnt = 0;
    for (int y=channel*len_right; y<(channel+1)*len_right; y++)
    {
      for (int x=len_lr-1; x>=0; x--)
      {
        tHSVColor hsv = hsv_pic[y*width + width-x-1];
        // perform calculations only if pixel has some luminosity (no black bars etc.)
        if (hsv.v > MinInputBrightness)
        {
          // find histogram position for pixel
          if ( (abs(most_used_hue[channel] - hsv.h) <= HistogramAnalysisWidth) ||
             (abs(most_used_hue[channel] - hsv.h) >= (360-HistogramAnalysisWidth)) )
          {
            // build histogram
            sat_hist[256*channel + hsv.s] += lr[x] * hsv.v;

            // build brightness average
            val_avg[channel] += hsv.v;
            val_cnt++;
          }
        }
      }
    }

    // calculate brightness average
    if (val_cnt > 0) { val_avg[channel] = val_avg[channel] / val_cnt; }
    else { val_avg[channel] = MinOutputBrightness; }
    // adjust brightness
    int new_value = (int) ((float)val_avg[channel] * ((float)BrightnessCorrection / 100.0));
    if (new_value > 255) { new_value = 255; } // ensure brightness isn't set too high
    val_avg[channel] = (unsigned char)new_value;
  }

  // Sat-Wert ermitteln rechts
  int* most_used_sat = new int[right];         // resulting sat for each channel
  memset(most_used_sat, 0, right*sizeof(int)); // clean

  for (int channel=0; channel<right; channel++)
  {
    int max = 0;
    for (int i = 0; i < 256; i++) // walk through histogram
    {
      int tmp = 0;
      for (int j = -HistogramAnalysisWidth; j < HistogramAnalysisWidth+1; j++)
      {
        int idx = i + j; // histogram item
        if (idx <   0) { idx = 256 - i; } // handle beginning of histogram -> roll back
        if (idx > 255) { idx = i; } // handle end of histogram -> roll forward

        tmp += sat_hist[256*channel + idx] * ((HistogramAnalysisWidth+1)-abs(j));
      }
      if (max < tmp)
      {
        max = tmp;
        most_used_sat[channel] = i;
      }
    }
  }

  /***************************************************************************/
  /* calculate results                                                       */
  /***************************************************************************/
  for (int channel = 0; channel < right; channel++) {
    col[channel] = HSV2RGB(most_used_hue[channel], most_used_sat[channel], val_avg[channel]); // convert back to rgb
  }

  delete [] most_used_hue;
  delete [] hue_hist;

  delete [] most_used_sat;
  delete [] sat_hist;

  delete [] val_avg;
}

void calc_top(tRGBColor *col)
{
  /***************************************************************************/
  /* Hue                                                                     */
  /***************************************************************************/
  // Hue-Histogramm ermitteln oben
  long int* hue_hist = new long int[360 * top];  // hue histogram
  memset(hue_hist, 0, 360*top*sizeof(long int)); // clean histogram

  for (int channel=0; channel<top; channel++) {
    for (int x=channel*len_top; x<(channel+1)*len_top; x++) {
      for (int y=0; y<len_tb; y++)
      {
        tHSVColor hsv = hsv_pic[y*width + x];
        // perform calculations only if pixel has some luminosity (no black bars etc.)
        if (hsv.v > MinInputBrightness)
        {
          // Add weight to channel
          hue_hist[360*channel + hsv.h] += tb[y] * hsv.v;
        }
      }
    }
  }

  // Hue-Wert ermitteln oben
  int* most_used_hue = new int[top];         // resulting hue for each channel
  memset(most_used_hue, 0, top*sizeof(int)); // clean

  for (int channel=0; channel<top; channel++)
  {
    int max = 0;
    for (int i = 0; i < 360; i++) // walk through histogram
    {
      int tmp = 0;
      for (int j = -HistogramAnalysisWidth; j < HistogramAnalysisWidth+1; j++)
      {
        int idx = i + j; // histogram item
        if (idx <   0) { idx = 360 - i; } // handle beginning of histogram -> roll back
        if (idx > 359) { idx = i; } // handle end of histogram -> roll forward

        tmp += hue_hist[360*channel + idx] * ((HistogramAnalysisWidth+1)-abs(j));
      }
      if (max < tmp)
      {
        max = tmp;
        most_used_hue[channel] = i;
      }
    }
  }

  /***************************************************************************/
  /* Saturation                                                              */
  /***************************************************************************/
  // Sat-Histogramm und Durchschnitts-Helligkeit (Value) ermitteln oben
  long int* sat_hist = new long int[256 * top];  // sat histogram
  memset(sat_hist, 0, 256*top*sizeof(long int)); // clean histogram

  int* val_avg = new int[top];         // resulting value for each channel
  memset(val_avg, 0, top*sizeof(int)); // clean

  for (int channel=0; channel<top; channel++)
  {
    int val_cnt = 0;
    for (int x=channel*len_top; x<(channel+1)*len_top; x++)
    {
      for (int y=0; y<len_tb; y++)
      {
        tHSVColor hsv = hsv_pic[y*width + x];
        // perform calculations only if pixel has some luminosity (no black bars etc.)
        if (hsv.v > MinInputBrightness)
        {
          // find histogram position for pixel
          if ( (abs(most_used_hue[channel] - hsv.h) <= HistogramAnalysisWidth) ||
             (abs(most_used_hue[channel] - hsv.h) >= (360-HistogramAnalysisWidth)) )
          {
            // build histogram
            sat_hist[256*channel + hsv.s] += tb[y] * hsv.v;

            // build brightness average
            val_avg[channel] += hsv.v;
            val_cnt++;
          }
        }
      }
    }

    // calculate brightness average
    if (val_cnt > 0) { val_avg[channel] = val_avg[channel] / val_cnt; }
    else { val_avg[channel] = MinOutputBrightness; }
    // adjust brightness
    int new_value = (int) ((float)val_avg[channel] * ((float)BrightnessCorrection / 100.0));
    if (new_value > 255) { new_value = 255; } // ensure brightness isn't set too high
    val_avg[channel] = (unsigned char)new_value;
  }

  // Sat-Wert ermitteln oben
  int* most_used_sat = new int[top];         // resulting sat for each channel
  memset(most_used_sat, 0, top*sizeof(int)); // clean

  for (int channel=0; channel<top; channel++)
  {
    int max = 0;
    for (int i = 0; i < 256; i++) // walk through histogram
    {
      int tmp = 0;
      for (int j = -HistogramAnalysisWidth; j < HistogramAnalysisWidth+1; j++)
      {
        int idx = i + j; // histogram item
        if (idx <   0) { idx = 256 - i; } // handle beginning of histogram -> roll back
        if (idx > 255) { idx = i; } // handle end of histogram -> roll forward

        tmp += sat_hist[256*channel + idx] * ((HistogramAnalysisWidth+1)-abs(j));
      }
      if (max < tmp)
      {
        max = tmp;
        most_used_sat[channel] = i;
      }
    }
  }

  /***************************************************************************/
  /* calculate results                                                       */
  /***************************************************************************/
  for (int channel = 0; channel < top; channel++) {
    col[channel] = HSV2RGB(most_used_hue[channel], most_used_sat[channel], val_avg[channel]); // convert back to rgb
  }

  delete [] most_used_hue;
  delete [] hue_hist;

  delete [] most_used_sat;
  delete [] sat_hist;

  delete [] val_avg;
}

void calc_bottom(tRGBColor *col)
{
  /***************************************************************************/
  /* Hue                                                                     */
  /***************************************************************************/
  // Hue-Histogramm ermitteln unten
  long int* hue_hist = new long int[360 * bottom];  // hue histogram
  memset(hue_hist, 0, 360*bottom*sizeof(long int)); // clean histogram

  for (int channel=0; channel<bottom; channel++) {
    for (int x=channel*len_bottom; x<(channel+1)*len_bottom; x++) {
      for (int y=0; y<len_tb; y++)
      {
        tHSVColor hsv = hsv_pic[(height-y-1)*width + x];
        // perform calculations only if pixel has some luminosity (no black bars etc.)
        if (hsv.v > MinInputBrightness)
        {
          // Add weight to channel
          hue_hist[360*channel + hsv.h] += tb[y] * hsv.v;
        }
      }
    }
  }

  // Hue-Wert ermitteln unten
  int* most_used_hue = new int[bottom];         // resulting hue for each channel
  memset(most_used_hue, 0, bottom*sizeof(int)); // clean

  for (int channel=0; channel<bottom; channel++)
  {
    int max = 0;
    for (int i = 0; i < 360; i++) // walk through histogram
    {
      int tmp = 0;
      for (int j = -HistogramAnalysisWidth; j < HistogramAnalysisWidth+1; j++)
      {
        int idx = i + j; // histogram item
        if (idx <   0) { idx = 360 - i; } // handle beginning of histogram -> roll back
        if (idx > 359) { idx = i; } // handle end of histogram -> roll forward

        tmp += hue_hist[360*channel + idx] * ((HistogramAnalysisWidth+1)-abs(j));
      }
      if (max < tmp)
      {
        max = tmp;
        most_used_hue[channel] = i;
      }
    }
  }

  /***************************************************************************/
  /* Saturation                                                              */
  /***************************************************************************/
  // Sat-Histogramm und Durchschnitts-Helligkeit (Value) ermitteln unten
  long int* sat_hist = new long int[256 * bottom];  // sat histogram
  memset(sat_hist, 0, 256*bottom*sizeof(long int)); // clean histogram

  int* val_avg = new int[bottom];         // resulting value for each channel
  memset(val_avg, 0, bottom*sizeof(int)); // clean

  for (int channel=0; channel<bottom; channel++)
  {
    int val_cnt = 0;
    for (int x=channel*len_bottom; x<(channel+1)*len_bottom; x++)
    {
      for (int y=0; y<len_tb; y++)
      {
        tHSVColor hsv = hsv_pic[(height-y-1)*width + x];
        // perform calculations only if pixel has some luminosity (no black bars etc.)
        if (hsv.v > MinInputBrightness)
        {
          // find histogram position for pixel
          if ( (abs(most_used_hue[channel] - hsv.h) <= HistogramAnalysisWidth) ||
             (abs(most_used_hue[channel] - hsv.h) >= (360-HistogramAnalysisWidth)) )
          {
            // build histogram
            sat_hist[256*channel + hsv.s] += tb[y] * hsv.v;

            // build brightness average
            val_avg[channel] += hsv.v;
            val_cnt++;
          }
        }
      }
    }

    // calculate brightness average
    if (val_cnt > 0) { val_avg[channel] = val_avg[channel] / val_cnt; }
    else { val_avg[channel] = MinOutputBrightness; }
    // adjust brightness
    int new_value = (int) ((float)val_avg[channel] * ((float)BrightnessCorrection / 100.0));
    if (new_value > 255) { new_value = 255; } // ensure brightness isn't set too high
    val_avg[channel] = (unsigned char)new_value;
  }

  // Sat-Wert ermitteln unten
  int* most_used_sat = new int[bottom];         // resulting sat for each channel
  memset(most_used_sat, 0, bottom*sizeof(int)); // clean

  for (int channel=0; channel<bottom; channel++)
  {
    int max = 0;
    for (int i = 0; i < 256; i++) // walk through histogram
    {
      int tmp = 0;
      for (int j = -HistogramAnalysisWidth; j < HistogramAnalysisWidth+1; j++)
      {
        int idx = i + j; // histogram item
        if (idx <   0) { idx = 256 - i; } // handle beginning of histogram -> roll back
        if (idx > 255) { idx = i; } // handle end of histogram -> roll forward

        tmp += sat_hist[256*channel + idx] * ((HistogramAnalysisWidth+1)-abs(j));
      }
      if (max < tmp)
      {
        max = tmp;
        most_used_sat[channel] = i;
      }
    }
  }

  /***************************************************************************/
  /* calculate results                                                       */
  /***************************************************************************/
  for (int channel = 0; channel < bottom; channel++) {
    col[bottom-channel-1] = HSV2RGB(most_used_hue[channel], most_used_sat[channel], val_avg[channel]); // convert back to rgb
  }

  delete [] most_used_hue;
  delete [] hue_hist;

  delete [] most_used_sat;
  delete [] sat_hist;

  delete [] val_avg;
}

// in: r, g, b: 0 - 255     out: h: 0 - 359; s, v: 0 - 255
tHSVColor RGB2HSV(uint8_t r, uint8_t g, uint8_t b)
{
  tHSVColor hsv;
  int max = r, min = r;

  // find maximum and minimum RGB values
  if (g > max)      max = g;
  else              min = g;
  if (b > max)      max = b;
  else if (b < min) min = b;

  hsv.v = max; // value
  const int delta = max - min;
  hsv.s = max ? (510*delta+max)/(2*max) : 0; // saturation

  // calc hue
  if (hsv.s == 0)
    hsv.h = 0; // undefined hue
  else if (r == max) // red
    hsv.h = (g >= b) ?   0 + (120*(g-b)+delta)/(2*delta) : 300 + (120*(g-b+delta)+delta)/(2*delta);
  else if (g == max) // green
    hsv.h = (b > r)  ? 120 + (120*(b-r)+delta)/(2*delta) :  60 + (120*(b-r+delta)+delta)/(2*delta);
  else // blue
    hsv.h = (r > g)  ? 240 + (120*(r-g)+delta)/(2*delta) : 180 + (120*(r-g+delta)+delta)/(2*delta);

  if (hsv.h == 360) { hsv.h = 0; }
  return hsv;
}

// in: h: 0 - 359; s, v: 0 - 255     out: r, g, b: 0 - 255
tRGBColor HSV2RGB(int h, int s, int v)
{
  tRGBColor rgb = {v, v, v};

  if (s != 0) // chromatic
  {
    unsigned int f = h % 60;
    h /= 60;
    unsigned int p = (2*v*(255-s)+255)/510;
    if (h & 1)
    {
      unsigned int q = (2*v*(15300-s*f)+15300)/30600;
      switch (h)
      {
        case 1: rgb.r = q; rgb.g = v; rgb.b = p; break;
        case 3: rgb.r = p; rgb.g = q; rgb.b = v; break;
        case 5: rgb.r = v; rgb.g = p; rgb.b = q; break;
      }
    }
    else
    {
      unsigned int t = (2*v*(15300-(s*(60-f)))+15300)/30600;
      switch (h)
      {
        case 0: rgb.r = v; rgb.g = t; rgb.b = p; break;
        case 2: rgb.r = p; rgb.g = v; rgb.b = t; break;
        case 4: rgb.r = t; rgb.g = p; rgb.b = v; break;
      }
    }
  }
  return rgb;
}
