/*
  Copyright 2009  Eike Edener (vdr [at] edener [dot] de)

  PB1: SCLK
  PB2: MOSI
  PB3: MISO
  PB6: XLAT
  PB7: GSCLK
  PB8: GND
  PB9: VCC

  PC5: BLANK
*/

/* Includes: */
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/iousb162.h>

#include "Descriptors.h"

#include <LUFA/Drivers/USB/USB.h>            /* USB Functionality                    */
#include <LUFA/Drivers/AT90USBXXX/SPI.h>     /* USART driver                         */
#include <LUFA/Scheduler/Scheduler.h>	     /* Simple scheduler for task management */

/* Macros: */
#define XLAT_LO         PORTB &= ~(1 << PB6)
#define XLAT_HI         PORTB |=  (1 << PB6)

#define TIMEOUT         36764 /* USB_MainTask is normally called every 13.6 us -> 13.6 us * 36764 = 500 ms */

/* Task Definitions: */
TASK(Init);
TASK(USB_MainTask);

/* Scheduler Task List: */
TASK_LIST
{
	{ Task: USB_USBTask , TaskStatus: TASK_STOP },
        { Task: Init        , TaskStatus: TASK_RUN  },
	{ Task: USB_MainTask, TaskStatus: TASK_STOP },
};

/* Event Handlers: */
HANDLES_EVENT(USB_Connect);
HANDLES_EVENT(USB_Reset);
HANDLES_EVENT(USB_Disconnect);
HANDLES_EVENT(USB_ConfigurationChanged);
HANDLES_EVENT(USB_UnhandledControlPacket);

/* Function Prototypes: */
void ConfigureSPI(void);

/* Global Variables: */
uint8_t dataToSend[IN_EP_SIZE];
uint8_t dataReceived[OUT_EP_SIZE];

uint8_t tempIOreg = 0;
uint8_t firstRun = 1;

uint16_t tlc_shift_reg_cnt = 0;
uint16_t spi_data_cnt = 0;
int8_t msg_idx = 0;

uint8_t need_latch = 0;

uint16_t timeout_cnt = 0;
uint16_t i = 0;

/** Main program entry point. This routine configures the hardware required by the application,
 *  then starts the scheduler to run the USB management task.
 */
int main(void)
{
	/* Disable watchdog if enabled by bootloader/fuses */
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	/* Disable Clock Division */
	SetSystemClockPrescaler(0);

        /* Disable Ports to avoid Linux debounce issues */
        DDRB = 0;
        PORTB = 0;
        DDRC = 0;
        PORTC = 0;
        DDRD = 0;
        PORTD = 0;

        /* SPI Initialization */
        ConfigureSPI();

        /* Prepare PortB for SPI - set PB0(^SS), PB1(SCK), PB2(MOSI) as output as well as all other pins except PB3(MISO) */
        DDRB = (1 << PB0) | (1 << PB1) | (1 << PB2) | (0 << PB3) | (1 << PB4) | (1 << PB5) | (1 << PB6) | (1 << PB7);
        PORTB |= (1 << PB0);
        XLAT_LO; /* no latch */

        /* Prepare PortC for Timer1 output */
        DDRC = (1 << PC5);
        PORTC |= (1 << PC5); /* blank all LEDs */

        /* Prepare PortD for Hardware-Button */
        PORTD |= (1 << PD7);

 	/* Initialize Scheduler so that it can be used */
	Scheduler_Init();

	/* Initialize USB Subsystem */
	USB_Init();

	/* Scheduling - routine never returns, so put this last in the main function */
	Scheduler_Start();
}

/** Configure SPI. */
void ConfigureSPI(void)
{
	uint8_t SPCRmask = (1 << SPE) | (1 << MSTR); /* enable SPI as Master */
	uint8_t SPSRmask = 0;

	/* MSB first */
	SPCRmask |= (0 << DORD);

	/* SPI-Mode 0: CPOL=0, CPHA=0: Sample on Rising Edge */
	SPCRmask |= ((0 << CPOL) | (0 << CPHA));

	/* Set speed: 4Mhz SPI (Fosc / 2) */
	SPCRmask |= ((0 << SPR1) | (0 << SPR0));
	SPSRmask |= (1 << SPI2X);

	SPCR = SPCRmask;
	SPSR = SPSRmask;

	/* only read if first run */
	if (firstRun)
	{
		tempIOreg = SPSR; /* need to read to initialize */
		tempIOreg = SPDR; /* need to read to initialize */
		firstRun = 0;
	}
}

/** Event handler for the USB_Connect event.
 *  Starts the library USB task to begin the enumeration and USB management process.
 */
EVENT_HANDLER(USB_Connect)
{
 	/* Start USB management task */
	Scheduler_SetTaskMode(USB_USBTask, TASK_RUN);
}

/** Event handler for the USB_Reset event. This fires when the USB interface is reset by the USB host, before the
 *  enumeration process begins, and enables the control endpoint interrupt so that control requests can be handled
 *  asynchronously when they arrive rather than when the control endpoint is polled manually.
 */
EVENT_HANDLER(USB_Reset)
{
	/* Select the control endpoint */
	Endpoint_SelectEndpoint(ENDPOINT_CONTROLEP);

	/* Enable the endpoint SETUP interrupt ISR for the control endpoint */
	USB_INT_Enable(ENDPOINT_INT_SETUP);
}

/** Event handler for the USB_Disconnect event.
 */
EVENT_HANDLER(USB_Disconnect)
{
	/* Stop running keyboard reporting and USB management tasks */
	Scheduler_SetTaskMode(USB_MainTask, TASK_STOP);
	Scheduler_SetTaskMode(USB_USBTask, TASK_STOP);
}

/** Event handler for the USB_ConfigurationChanged event. This is fired when the host sets the current configuration
 *  of the USB device after enumeration, and configures the keyboard device endpoints.
 */
EVENT_HANDLER(USB_ConfigurationChanged)
{
	/* Setup Keyboard Keycode Report Endpoint */
	Endpoint_ConfigureEndpoint(IN_EP, EP_TYPE_INTERRUPT,
				   ENDPOINT_DIR_IN, IN_EP_SIZE,
				   ENDPOINT_BANK_SINGLE);

	/* Enable the endpoint IN interrupt ISR for data being sent TO the host */
	USB_INT_Enable(ENDPOINT_INT_IN);

	/* Setup Keyboard LED Report Endpoint */
	Endpoint_ConfigureEndpoint(OUT_EP, EP_TYPE_INTERRUPT,
				   ENDPOINT_DIR_OUT, OUT_EP_SIZE,
				   ENDPOINT_BANK_SINGLE);

	/* Enable the endpoint OUT interrupt ISR for data recevied FROM the host */
	USB_INT_Enable(ENDPOINT_INT_OUT);

	/* start USB_USBTask */
	Scheduler_SetTaskMode(USB_MainTask, TASK_RUN);
}

/** Event handler for the USB_UnhandledControlPacket event. This is used to catch standard and class specific
 *  control requests that are not handled internally by the USB library (including the HID commands, which are
 *  all issued via the control endpoint), so that they can be handled appropriately for the application.
 */
EVENT_HANDLER(USB_UnhandledControlPacket)
{
	/* Handle HID Class specific requests here (these are Control EP requests) */
	switch (bRequest)
	{
		case 0x01:
			if (bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				Endpoint_ClearSetupReceived();

				/* Write the report data to the control endpoint */
				Endpoint_Write_Control_Stream_LE(&dataToSend, sizeof(dataToSend));

				/* Finalize the transfer, acknowedge the host error or success OUT transfer */
				Endpoint_ClearSetupOUT();
			}
		break;
	}
}

/** USB ISR */
ISR(ENDPOINT_PIPE_vect)
{
	/* Check if the control endpoint has received a request */
	if (Endpoint_HasEndpointInterrupted(ENDPOINT_CONTROLEP))
	{
		/* Clear the endpoint interrupt */
		Endpoint_ClearEndpointInterrupt(ENDPOINT_CONTROLEP);

		/* Process the control request */
		USB_USBTask();

		/* Handshake the endpoint setup interrupt - must be after the call to USB_USBTask() */
		USB_INT_Clear(ENDPOINT_INT_SETUP);
	}

	/* Check if IN endpoint has interrupted (data needs to be sent to host) */
	if (Endpoint_HasEndpointInterrupted(IN_EP))
	{
		/* Select the endpoint */
		Endpoint_SelectEndpoint(IN_EP);

		/* Clear the endpoint IN interrupt flag */
		USB_INT_Clear(ENDPOINT_INT_IN);

		/* Clear the IN endpoint interrupt */
		Endpoint_ClearEndpointInterrupt(IN_EP);

		/* Send Data over USB */
		Endpoint_Write_Stream_LE(&dataToSend, IN_EP_SIZE);

		/* Handshake the IN Endpoint - send the data to the host */
		Endpoint_ClearCurrentBank();
	}

	/* Check if the OUT Endpoint has interrupted (received data from the host) */
	if (Endpoint_HasEndpointInterrupted(OUT_EP))
	{
		/* Select the endpoint */
		Endpoint_SelectEndpoint(OUT_EP);

		/* Clear the endpoint OUT interrupt flag */
		USB_INT_Clear(ENDPOINT_INT_OUT);

		/* Clear the OUT endpoint interrupt */
		Endpoint_ClearEndpointInterrupt(OUT_EP);

		/* Read in data from the host */
		if (0 == Endpoint_Read_Stream_LE(&dataReceived, 61))
		{
			if (msg_idx == dataReceived[0])
			{
				for (i=1; i<61; i++)
				{
                        		/* Send data over SPI */
	                        	SPI_SendByte(dataReceived[i]);
        	                	spi_data_cnt++;
					if (spi_data_cnt >= tlc_shift_reg_cnt)
					{
                                        	need_latch = 1;
						spi_data_cnt = 0;
						msg_idx = -1; /* increment comes next :-) */

						break;
					}
				}

	                        /* increment message index */
        	                msg_idx++;

                	        /* Reset timeout */
				timeout_cnt = 0;
			}
		}

		/* Handshake the OUT Endpoint - clear endpoint and ready for next report */
		Endpoint_ClearCurrentBank();
	}

	/* remove any pending Timer 1 Compare-B interrupt */
	TIFR1 = 0x04;
}

/** Timer 1 Compare-B ISR */
ISR(TIMER1_COMPB_vect)
{
	if (need_latch)
	{
		/* Latch new data */
		XLAT_HI;
		XLAT_LO;
		need_latch = 0;
	}
}

/** Task for initialisation. */
TASK(Init)
{
	/* initialize the send and receive buffers */
	for (i = 0; i <  IN_EP_SIZE; i++) { dataToSend[i] = 0; }
	for (i = 0; i < OUT_EP_SIZE; i++) { dataReceived[i] = 0; }

	/* fill all possibly available TLCs (max. 40)
         *  - SPI_SendByte blocks here, when ring is open! */
	for (i=0; i<40*24; i++) { SPI_SendByte(0xFF); }

	/* Fill all TLCs with Zeros and wait for the first coming back */
	while (tlc_shift_reg_cnt < (40*24+1))
	{
		tempIOreg = SPI_TransferByte(0x00);
		if (tempIOreg == 0) break;
		tlc_shift_reg_cnt++;
	}

        /* SPI ring length */
	dataToSend[0] = tlc_shift_reg_cnt;
	dataToSend[1] = (tlc_shift_reg_cnt >> 8);

	/* Latch new data */
	XLAT_HI;
	XLAT_LO;

	/* Prepare Timer0 for Greyscale Clock: Mode 2 (CTC), no prescaling, 500 kHz, output compare PB7 */
	TCCR0A = (0 << COM0A1) | (1 << COM0A0) | (0 << COM0B1) | (0 << COM0B0) | (1 << WGM01) | (0 << WGM00);
	TCCR0B = (0 << FOC0A) | (0 << FOC0B) |  (0 << WGM02) | (0 << CS02) | (0 << CS01) | (1 << CS00);
	OCR0A  = 7;

	/* Prepare Timer1 for BLANK: Mode 15 (Fast PWM), no prescaling, 122 Hz, output compare PC6 */
	TCCR1A = (0 << COM1A1) | (0 << COM1A0) | (1 << COM1B1) | (1 << COM1B0) | (0 << COM1C1) | (0 << COM1C0) | (1 << WGM11) | (1 << WGM10);
	TCCR1B = (0 << ICNC1) | (0 << ICES1) | (1 << WGM13) | (1 << WGM12) | (0 << CS12) | (0 << CS11) | (1 << CS10);
	OCR1A  = 65535; /* this is for 122 Hz (data sheet says: 32767 = 4096 * 8 - 1 !?) */
	OCR1B  = 65508;

	/* reset timer counter */
	TCNT0 = 0;
	TCNT1 = 0;

        /* enable interrupt */
        TIMSK1 = (1 << OCIE1B);

        /* start the watchdog */
	wdt_enable(WDTO_8S);

	/* run only once */
	Scheduler_SetTaskMode(Init, TASK_STOP);
}

/** Main Task */
TASK(USB_MainTask)
{
	/* Reset the watchdog timer */
	wdt_reset();

	/* Check if the USB System is connected to a Host */
	if (USB_IsConnected)
	{
		/* process data or do something generally useful */
		/* note that TASK(USB_MainTask) will be periodically executed when no other tasks or functions are running */

                timeout_cnt++;

                /* Check timeout */
                if (timeout_cnt >= TIMEOUT)
                {
			timeout_cnt  = 0;
			spi_data_cnt = 0;
 			msg_idx      = 0;
                }
	}
}
