/*
 * outputusb.c: USB output
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <usb.h>

#include <vdr/tools.h>
#include <vdr/thread.h>

#include "defs.h"
#include "setup.h"
#include "output.h"
#include "outputusb.h"

#define INTERFACENUM  0

static usb_dev_handle *dev_handle;

static char *output;
static char send[61];

static uint16_t ring_length_char;
static uint16_t ring_length_12bit;

static uint16_t gamma_tab_r[256], gamma_tab_g[256], gamma_tab_b[256];

static void setLED(uint8_t lednr, uint16_t intensity)
{
  uint16_t bitnr = lednr * 12; // 12 bits per led
  uint8_t bytenr = bitnr >> 3;
  bool startnibble = (bitnr & 7) ? true : false;

  if (!startnibble)
  {
    output[bytenr] = intensity >> 4;
    output[bytenr+1] = (output[bytenr+1] & 0x0f) | ((intensity & 0x0f) << 4);
  }
  else
  {
    output[bytenr] = (output[bytenr] & 0xf0) | (intensity >> 8);
    output[bytenr+1] = intensity & 0xff;
  }
}

static void setRGB(uint8_t channel, uint16_t r, uint16_t g, uint16_t b)
{
  setLED(ring_length_12bit - (3*channel+0) - 1, r);
  setLED(ring_length_12bit - (3*channel+1) - 1, g);
  setLED(ring_length_12bit - (3*channel+2) - 1, b);
}

static int usbGetStringAscii(usb_dev_handle *dev, int index, int langid, char *buf, int buflen)
{
  char buffer[256];
  int  rval, i;

  if ((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, buffer, sizeof(buffer), 1000)) < 0)
     return rval;
  if (buffer[1] != USB_DT_STRING)
     return 0;
  if ((unsigned char)buffer[0] < rval)
     rval = (unsigned char)buffer[0];
  rval /= 2;

  // lossy conversion to ISO Latin1
  for (i=1;i<rval;i++)
  {
    if (i > buflen) // destination buffer overflow
       break;
    buf[i-1] = buffer[2 * i];
    if (buffer[2 * i + 1] != 0) // outside of ISO Latin1 range
       buf[i-1] = '?';
  }
  buf[i-1] = 0;

  return (i-1);
}

static void find_usb_device(void)
{
  struct usb_bus *bus;
  struct usb_device *dev;

  char manufacturer[30];
  char product[30];
  char serial_number[30];

  usb_init();
  usb_find_busses();
  usb_find_devices();

  for (bus = usb_get_busses(); bus; bus = bus->next)
  {
    for (dev = bus->devices; dev; dev = dev->next)
    {
      if ((dev->descriptor.idVendor == 0x03EB) && (dev->descriptor.idProduct == 0x204F))
      {
        int len;

        dev_handle = usb_open(dev);
        if(!dev_handle)
        {
          esyslog("Solarstorm: cannot open USB device: %s\n", usb_strerror());
          continue;
        }

        len = usbGetStringAscii(dev_handle, dev->descriptor.iProduct, 0x0409, product, sizeof(product));
        if ((len > 0) && (NULL != strstr(product, "Solarstorm")))
        {
          esyslog("Solarstorm: USB-product %s found\n", product);
        }
        else
        {
          esyslog("Solarstorm: wrong USB-product string");
          /* make sure other programs can still access this device */
          /* release the interface and close the device */
          usb_release_interface(dev_handle, INTERFACENUM);
          usb_close(dev_handle);

          dev_handle = NULL;
          continue;
        }

        len = usbGetStringAscii(dev_handle, dev->descriptor.iManufacturer, 0x0409, manufacturer, sizeof(manufacturer));
        if (len > 0) esyslog("Solarstorm: USB-manufacturer %s found\n", manufacturer);

        len = usbGetStringAscii(dev_handle, dev->descriptor.iSerialNumber, 0x0409, serial_number, sizeof(serial_number));
        if (len > 0) esyslog("Solarstorm: serial number %s found \n", serial_number);

        /* tell libusb to use the 1. configuration of the device */
        if (0 < usb_set_configuration(dev_handle, 1))
        {
          esyslog("Solarstorm: could not set configuration! No permissions?");
          usb_close(dev_handle);
          dev_handle = NULL;
          continue;
        }

        #ifdef LIBUSB_HAS_DETACH_KERNEL_DRIVER_NP
        /* detach the automatically assigned kernel driver from the current interface */
        usb_detach_kernel_driver_np(dev_handle, INTERFACENUM);
        #endif

        /* claim the interface for use by this program */
        if (0 < usb_claim_interface(dev_handle, INTERFACENUM))
        {
          esyslog("Solarstorm: could not claim interface! No permissions?");
          usb_close(dev_handle);
          dev_handle = NULL;
          continue;
        }

        return;
      }
    }
  }
}

cOutputUSB::cOutputUSB(unsigned int sum)
{
  cnt = sum;
  dev_handle = NULL;
  output = NULL;
  ring_length_char = 0;
  ring_length_12bit = 0;

  CalcGamma();
}

void cOutputUSB::CalcGamma(void)
{
  double GammaR = (double)SolarstormSetup.Gamma.r / 100.0f;
  double GammaG = (double)SolarstormSetup.Gamma.g / 100.0f;
  double GammaB = (double)SolarstormSetup.Gamma.b / 100.0f;

  for (int i=0; i<256; i++) 
  {
    gamma_tab_r[i] = (uint16_t)(pow((double)i / 255.0f, GammaR) * 4095.0f);
    gamma_tab_g[i] = (uint16_t)(pow((double)i / 255.0f, GammaG) * 4095.0f);
    gamma_tab_b[i] = (uint16_t)(pow((double)i / 255.0f, GammaB) * 4095.0f);
  }
}

bool cOutputUSB::Open(const char *param)
{
  char buffer[8];

  find_usb_device();
  if (!dev_handle)
  {
    esyslog("Solarstorm: No device found!");
    return false;
  }

  int ret = usb_interrupt_read(dev_handle, 0x81, buffer, 8, 500); // receive with 500 ms timeout
  if (ret >= 2)
  {
    ring_length_char  = buffer[1]*256 + buffer[0];
    ring_length_12bit = ring_length_char * 8 / 12;
    if (ring_length_12bit > 0)
    {
      esyslog("Solarstorm: ring length = %d outputs\n", ring_length_12bit);
    }
    else
    {
      esyslog("Solarstorm: ring length = 0! Device jumpered?\n");
      return false;
    }
  }
  else
  {
    esyslog("Solarstorm: Device did not send ring length!\n");
    return false;
  }

  output = new char[((ring_length_char / 60) + 1) * 60 + 1];

  return true;
}

bool cOutputUSB::Close()
{
  if (dev_handle)
  {
    /* switch off */
    int msg = (ring_length_char / 60) + 1;
    for (int i = 0; i < msg; i++)
    {
      send[0] = i;
      for (int j = 1; j < 61; j++)
      {
        send[j] = 0;
      }
      usb_interrupt_write(dev_handle, 0x02, send, 61, 500); // send with 500 ms timeout
    }

    /* make sure other programs can still access this device */
    /* release the interface and close the device */
    usb_release_interface(dev_handle, INTERFACENUM);
    usb_close(dev_handle);
    delete[] output;
    dev_handle = NULL;
    output = NULL;
    ring_length_char = 0;
    ring_length_12bit = 0;
    return true;
  }
  else
  {
    delete[] output;
    output = NULL;
    ring_length_char = 0;
    ring_length_12bit = 0;
    return false;
  }

  return false;
}

bool cOutputUSB::OutputColors(tRGBColor *col)
{
  for (unsigned int i = 0; i < cnt; i++) // l, t, r, b
  {
    // white calibration
    col[i].r = (unsigned char)((int)SolarstormSetup.WhiteCalibration[i].r * (int)col[i].r / 255);
    col[i].g = (unsigned char)((int)SolarstormSetup.WhiteCalibration[i].g * (int)col[i].g / 255);
    col[i].b = (unsigned char)((int)SolarstormSetup.WhiteCalibration[i].b * (int)col[i].b / 255);   

    // output with gamma correction
    setRGB(i, gamma_tab_r[col[i].r], gamma_tab_g[col[i].g], gamma_tab_b[col[i].b]);
  }

  // send data to target
  if (dev_handle)
  {
    int msg = (ring_length_char / 60) + 1;
    for (int i = 0; i < msg; i++)
    {
      send[0] = i;
      memcpy(&send[1], &output[60*i], 60);
      usb_interrupt_write(dev_handle, 0x02, send, 61, 500); // send with 500 ms timeout
    }
    return true;
  }
  else
  {
    return false;
  }
}
