/*
 * solarstorm.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <getopt.h>
#include <stdlib.h>
#include <vdr/plugin.h>

#include "defs.h"
#include "menu.h"
#include "setup.h"
#include "thread.h"
#include "whitecalibration.h"
#include "solarstorm.h"

cPlugin *SolarstormPlugin;

cSolarstormSetup SolarstormSetup;
cWorkThread *WorkThread;

tInput  InputType  = grab; // input device type
tOutput OutputType = usb;  // output device type

bool WhiteCalibration = false; // white calibration on/off

cPluginSolarstorm::cPluginSolarstorm()
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!

  SolarstormPlugin = this;
  WorkThread = NULL;
}

cPluginSolarstorm::~cPluginSolarstorm()
{
  // Clean up after yourself!
}

void cPluginSolarstorm::ResetFilter() { WorkThread->ResetFilter(); }
void cPluginSolarstorm::CalcGamma() { WorkThread->CalcGamma(); }
void cPluginSolarstorm::DoWhiteCalibration() { WhiteCalibration = true; }

const char *cPluginSolarstorm::CommandLineHelp(void)
{
  // Return a string that describes all known command line options.
  return "  no command line options available";
}

bool cPluginSolarstorm::ProcessArgs(int argc, char *argv[])
{
  // Implement command line argument processing here if applicable.
  return true;
}

bool cPluginSolarstorm::Initialize(void)
{
  // Initialize any background activities the plugin shall perform.
  return true;
}

bool cPluginSolarstorm::Start(void)
{
  // Start any background activities the plugin shall perform.
  cWhiteCalibration::Init();
  WorkThread = new cWorkThread(InputType, OutputType);
  return (WorkThread ? true : false);
}

void cPluginSolarstorm::Stop(void)
{
  // Stop any background activities the plugin shall perform.
  delete WorkThread;
  WorkThread = NULL;
  cWhiteCalibration::Exit();
}

void cPluginSolarstorm::Housekeeping(void)
{
  // Perform any cleanup or other regular tasks.
}

void cPluginSolarstorm::MainThreadHook(void)
{
  // Perform actions in the context of the main program thread.
  // WARNING: Use with great care - see PLUGINS.html!
}

cString cPluginSolarstorm::Active(void)
{
  // Return a message string if shutdown should be postponed
  return NULL;
}

time_t cPluginSolarstorm::WakeupTime(void)
{
  // Return custom wakeup time for shutdown script
  return 0;
}

const char *cPluginSolarstorm::MainMenuEntry(void)
{
  if ((no_input == InputType) ||
      (no_output == OutputType))
  {
    return NULL;
  }

  if (WorkThread->Active())
    return tr("Solarstorm off");
  else
    return tr("Solarstorm on");
}

cOsdObject *cPluginSolarstorm::MainMenuAction(void)
{
  // Perform the action when selected from the main VDR menu.
  if (WhiteCalibration)
  {
    WhiteCalibration = false;
    return new cWhiteCalibration;
  }

  WorkThread->Active() ? WorkThread->SwitchOff() : WorkThread->SwitchOn();

  return NULL;
}

cMenuSetupPage *cPluginSolarstorm::SetupMenu(void)
{
  // Return a setup menu in case the plugin supports one.
  return new cMenuSetupMenu;
}

bool cPluginSolarstorm::SetupParse(const char *Name, const char *Value)
{
  // Parse your own setup parameters and store their values.
       if (cMenuSetupMenu::SetupParse(Name, Value)) ;
  else if (cMenuSetupMode::SetupParse(Name, Value)) ;
  else if (cMenuSetupFilter::SetupParse(Name, Value)) ;
  else if (cMenuSetupChannels::SetupParse(Name, Value)) ;
  else if (cMenuSetupGamma::SetupParse(Name, Value)) ;
  else if (cWhiteCalibration::SetupParse(Name, Value)) ;
  else return false;

  return true;
}

bool cPluginSolarstorm::Service(const char *Id, void *Data)
{
  // Handle custom service requests from other plugins
  return false;
}

const char **cPluginSolarstorm::SVDRPHelpPages(void)
{
  // Return help text for SVDRP commands this plugin implements
  static const char *HelpPages[] = {
    "ON\n"
    "    Switch Solarstorm on.",
    "OFF\n"
    "    Switch Solarstorm off.",
    "STATUS\n"
    "    Print Solarstorm status.",
    NULL
  };
  return HelpPages;
}

cString cPluginSolarstorm::SVDRPCommand(const char *Command, const char *Option, int &ReplyCode)
{
  // Process SVDRP commands this plugin implements

  if (strcasecmp(Command, "ON") == 0)
  {
    if (!WorkThread->Active()) { WorkThread->SwitchOn(); }
    return cString::sprintf("Solarstorm: ON");
  }
  else if (strcasecmp(Command, "OFF") == 0)
  {
    if (WorkThread->Active()) { WorkThread->SwitchOff(); }
    return cString::sprintf("Solarstorm: OFF");
  }
  else if (strcasecmp(Command, "STATUS") == 0)
  {
    if (WorkThread->Active())
    {
      return cString::sprintf("Solarstorm: ON");
    }
    else
    {
      return cString::sprintf("Solarstorm: OFF");
    }
  }
  else
  {
    return NULL;
  }

  return NULL; // no compiler warning
}

VDRPLUGINCREATOR(cPluginSolarstorm); // Don't touch this!
