/*
 * thread.c: Work thread
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */
#include <stdlib.h>
#include <vdr/skins.h>
#include <vdr/tools.h>
#include <vdr/thread.h>

#include "math.h"
#include "defs.h"

// calculation class
#include "calculation.h"

// filter class
#include "filter.h"

// input classes
#include "input.h"
#include "inputgrab.h"

// output classes
#include "output.h"
#include "outputusb.h"

#include "setup.h"
#include "thread.h"

// seconds the next timer must be in the future to assume manual start
#define MANUALSTART       600

static cInput     *Input;  // input device
static cCalc      *Calc;   // calculation
static cColFilter *Filter; // filter
static cOutput    *Output; // output device

static tRGBColor *color;
static tHSVColor *HSV_Img;

static unsigned char left, top, right, bottom, sum;

cWorkThread::cWorkThread(tInput InputType, tOutput OutputType):cThread("Solarstorm-Thread")
{
  left   = SolarstormSetup.ChannelsLeft;
  top    = SolarstormSetup.ChannelsTop;
  right  = SolarstormSetup.ChannelsRight;
  bottom = SolarstormSetup.ChannelsBottom;
  sum    = left + top + right + bottom;

  color = new tRGBColor[sum];
  HSV_Img = new tHSVColor[SolarstormSetup.CaptureWidth * SolarstormSetup.CaptureHeight];
  if ((NULL == color) || (NULL == HSV_Img))
  {
    Skins.QueueMessage(mtInfo, tr("Solarstorm: No memory available!"));
  }

  // input device
  switch (InputType)
  {
    case grab:
         Input = new cInputGrab(SolarstormSetup.CaptureWidth, SolarstormSetup.CaptureHeight);
    break;

    default:
    case no_input:
         Input = NULL;
         Skins.QueueMessage(mtInfo, tr("Solarstorm: No input device given!"));
    break;
  }

  // calculation
  Calc = new cCalc(left, top, right, bottom, SolarstormSetup.CaptureWidth, SolarstormSetup.CaptureHeight);

  // filter
  Filter = new cColFilter(sum);

  // output device
  switch (OutputType)
  {
    case usb:
         Output = new cOutputUSB(sum);
    break;

    default:
    case no_output:
         Output = NULL;
         Skins.QueueMessage(mtInfo, tr("Solarstorm: No output device given!"));
    break;
  }

  FirstStart();
}

cWorkThread::~cWorkThread()
{
  if (Running())
  {
    Cancel(1);
  }

  if (Input)
  {
    Input->Close();
    delete Input;
    Input = NULL;
  }

  if (Calc)
  {
    delete Calc;
    Calc = NULL;
  }

  if (Filter)
  {
    delete Filter;
    Filter = NULL;
  }

  if (Output)
  {
    Output->Close();
    delete Output;
    Output = NULL;
  }

  delete[] color;
  delete[] HSV_Img;
}

void cWorkThread::FirstStart(void)
{
  if ((NULL == Input) || (NULL == Output) || (NULL == color)) { return; } // input and output existing?

  cCondWait::SleepMs(1000); // wait 1000ms for loaded timers

  // manual or automatic start of vdr?
       if (SolarstormSetup.StartMode == 0) { SwitchOn(); } // on
  else if (SolarstormSetup.StartMode == 1) { SwitchOff(); } // off
  else if (SolarstormSetup.StartMode == 2) // timer dependent
  {
    time_t Now = time(NULL);
    cTimer *timer = Timers.GetNextActiveTimer();
    time_t Next  = timer ? timer->StartTime() : 0;
    time_t Delta = timer ? Next - Now : 0;

    if (!timer || Delta > MANUALSTART)
    { // Apparently the user started VDR manually
      SwitchOn();
    }
    else
    {
      SwitchOff();
    }
  }
}

void cWorkThread::SwitchOn(void)
{
  if (false == Input->Open(NULL))
  {
    Skins.QueueMessage(mtInfo, tr("Solarstorm: Could not open input device!"));
    return;
  }
  if (false == Output->Open(NULL))
  {
    Skins.QueueMessage(mtInfo, tr("Solarstorm: Could not open output device!"));
    return;
  }
  Filter->ResetFilter();
  Start();
}

void cWorkThread::SwitchOff(void)
{
  Cancel(1);
  Input->Close();
  Output->Close();
}

void cWorkThread::ResetFilter(void)
{
  Filter->ResetFilter();
}

void cWorkThread::CalcGamma(void)
{
  Output->CalcGamma();
}

void cWorkThread::Action(void)
{
  cTimeMs t;

  while (Running())
  {
    t.Set(0);

    // get calculated color from input
    if (Input->GetHSVImg(HSV_Img))
    {
      cCondWait::SleepMs(1);

      // calc colors
      Calc->CalcChannels(color, HSV_Img);

      // send to filter
      Filter->DoFilter(color);

      // send to output device
      OutputColors(color);
    }

    if (t.Elapsed() < 20) { cCondWait::SleepMs(20 - t.Elapsed()); }
  }
}

// send filter-output to output-device
void cWorkThread::OutputColors(tRGBColor *col)
{
  if (static_color == SolarstormSetup.Mode)
  {
    switch (SolarstormSetup.StaticColor)
    {
      case black:
      default:
           for (int i=0; i<sum; i++)
           {
             col[i].r = 0; // red (0-255)
             col[i].g = 0; // green (0-255)
             col[i].b = 0; // blue (0-255)
           }
      break;

      case white:
           for (int i=0; i<sum; i++)
           {
             col[i].r = 255; // red (0-255)
             col[i].g = 255; // green (0-255)
             col[i].b = 255; // blue (0-255)
           }
      break;

      case red:
           for (int i=0; i<sum; i++)
           {
             col[i].r = 255; // red (0-255)
             col[i].g = 0;   // green (0-255)
             col[i].b = 0;   // blue (0-255)
           }
      break;

      case green:
           for (int i=0; i<sum; i++)
           {
             col[i].r = 0;   // red (0-255)
             col[i].g = 255; // green (0-255)
             col[i].b = 0;   // blue (0-255)
           }
      break;

      case blue:
           for (int i=0; i<sum; i++)
           {
             col[i].r = 0;   // red (0-255)
             col[i].g = 0;   // green (0-255)
             col[i].b = 255; // blue (0-255)
           }
      break;

      case yellow:
           for (int i=0; i<sum; i++)
           {
             col[i].r = 255; // red (0-255)
             col[i].g = 255; // green (0-255)
             col[i].b = 0;   // blue (0-255)
           }
      break;

      case cyan:
           for (int i=0; i<sum; i++)
           {
             col[i].r = 0;   // red (0-255)
             col[i].g = 255; // green (0-255)
             col[i].b = 255; // blue (0-255)
           }
      break;

      case magenta:
           for (int i=0; i<sum; i++)
           {
             col[i].r = 255; // red (0-255)
             col[i].g = 0;   // green (0-255)
             col[i].b = 255; // blue (0-255)
           }
      break;
    }
  }

  Output->OutputColors(col);
}
