/*
 * whitecalibration.c: White calibration
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id$
 */

#include <vdr/device.h>
#include <vdr/plugin.h>

#include "defs.h"
#include "setup.h"
#include "whitecalibration.h"

#define MAX_LINES	4 // number of lines

#define WHITECALIBRATION	"WhiteCalibration"

static const tColor BCOLOR = clrBlack;   // black
static const tColor FCOLOR = 0xFFFFFFFF; // white

static char *parameter = NULL;

extern cPlugin *SolarstormPlugin;

cWhiteCalibration::cWhiteCalibration(void)
{
  osd = NULL;

#if APIVERSNUM >= 10708
  double aspect;
  cDevice::PrimaryDevice()->GetOsdSize(osd_width, osd_height, aspect);
  osd_width--;
  osd_height--;
#else
  osd_height = 575;
  osd_width  = 719;
#endif

  cur_line = 1;
  cur_channel = 1;

  old_Mode = SolarstormSetup.Mode;
  old_FilterMode = SolarstormSetup.FilterMode;

  SolarstormSetup.Mode = live_picture_hsv;
  SolarstormSetup.FilterMode = no_filter;
}

cWhiteCalibration::~cWhiteCalibration()
{
  SolarstormSetup.Mode = old_Mode;
  SolarstormSetup.FilterMode = old_FilterMode;
  delete osd;
}

void cWhiteCalibration::Init(void)
{
  char *par = parameter;
  int i;
  int sum = SolarstormSetup.ChannelsLeft + SolarstormSetup.ChannelsRight + SolarstormSetup.ChannelsTop + SolarstormSetup.ChannelsBottom;
  SolarstormSetup.WhiteCalibration = new tRGBColor[sum];

  for (i=0; i<sum; i++)
  {
    SolarstormSetup.WhiteCalibration[i].r = SolarstormSetup.WhiteCalibration[i].g = SolarstormSetup.WhiteCalibration[i].b = 255;
  }

  if (parameter == NULL) return;

  i = 0;
  while (1)
  {
    char *ptr;
    int value;

    value = strtol(par, &ptr, 10);
    if (par == ptr || i >= sum) break;
    SolarstormSetup.WhiteCalibration[i].r = value;
    par = ptr;

    value = strtol(par, &ptr, 10);
    if (par == ptr || i >= sum) break;
    SolarstormSetup.WhiteCalibration[i].g = value;
    par = ptr;

    value = strtol(par, &ptr, 10);
    if (par == ptr || i >= sum) break;
    SolarstormSetup.WhiteCalibration[i++].b = value;
    par = ptr;
  }

  delete [] parameter;
}

void cWhiteCalibration::Exit(void)
{
  delete [] SolarstormSetup.WhiteCalibration;
}

void cWhiteCalibration::AdjustChannel(int step)
{
  int sum = SolarstormSetup.ChannelsLeft + SolarstormSetup.ChannelsRight + SolarstormSetup.ChannelsTop + SolarstormSetup.ChannelsBottom;
  if ((cur_channel + step >= 1) && (cur_channel + step <= sum))
  {
    cur_channel += step;
  }
}

void cWhiteCalibration::AdjustValue(unsigned char *value, int step)
{
  if ((*value + step >= 0) && (*value + step <= 255)) { *value += step; }
}

void cWhiteCalibration::Adjust(int step)
{
  switch (cur_line)
  {
    case 1: AdjustChannel(step); break;
    case 2: AdjustValue(&SolarstormSetup.WhiteCalibration[cur_channel-1].r, step); break;
    case 3: AdjustValue(&SolarstormSetup.WhiteCalibration[cur_channel-1].g, step); break;
    case 4: AdjustValue(&SolarstormSetup.WhiteCalibration[cur_channel-1].b, step); break;
  }
}

void cWhiteCalibration::DrawMark(void)
{
  if (cur_channel <= SolarstormSetup.ChannelsLeft)
  {
    int len = (osd_height+1) / SolarstormSetup.ChannelsLeft;
    int pos = (osd_height+1) - (cur_channel * len) + (len/2) - (cFont::GetFont(fontOsd)->Height() / 2);
    osd->DrawText(20, pos, "X", BCOLOR, FCOLOR, cFont::GetFont(fontOsd));
  }
  else if (cur_channel <= SolarstormSetup.ChannelsLeft + SolarstormSetup.ChannelsTop)
  {
    int len = (osd_width+1) / SolarstormSetup.ChannelsTop;
    int pos = ((cur_channel - SolarstormSetup.ChannelsLeft - 1) * len) + (len/2);
    osd->DrawText(pos, 20, "X", BCOLOR, FCOLOR, cFont::GetFont(fontOsd));  
  }
  else if (cur_channel <= SolarstormSetup.ChannelsLeft + SolarstormSetup.ChannelsTop + SolarstormSetup.ChannelsRight)
  {
    int len = (osd_height+1) / SolarstormSetup.ChannelsRight;
    int pos = ((cur_channel - SolarstormSetup.ChannelsLeft - SolarstormSetup.ChannelsTop - 1) * len) + (len/2) - (cFont::GetFont(fontOsd)->Height() / 2);
    osd->DrawText(osd_width - 19 - cFont::GetFont(fontOsd)->Width("X"), pos, "X", BCOLOR, FCOLOR, cFont::GetFont(fontOsd));
  }
  else 
  {
    int len = (osd_width+1) / SolarstormSetup.ChannelsBottom;
    int pos = (osd_width+1) - ((cur_channel - SolarstormSetup.ChannelsLeft - SolarstormSetup.ChannelsTop - SolarstormSetup.ChannelsRight) * len) + (len/2);
    osd->DrawText(pos, osd_height - 19 - cFont::GetFont(fontOsd)->Height(), "X", BCOLOR, FCOLOR, cFont::GetFont(fontOsd));  
  }
}

void cWhiteCalibration::DrawLine(int line_nr, const char *desc, int value)
{
  char tmp[25];
  tColor fc, bc;
  int length;
  int font_height = 2 * cFont::GetFont(fontOsd)->Height();
  int vert_start = (osd_height - (MAX_LINES * font_height)) / 2 - font_height;

  if (cur_line == line_nr) { bc = BCOLOR; fc = FCOLOR; } else { bc = FCOLOR; fc = BCOLOR; }

  sprintf(tmp, " %s ", desc);
  length = cFont::GetFont(fontOsd)->Width(tmp);
  osd->DrawText(osd_width / 2 - length - 20, vert_start + line_nr * font_height, tmp, fc, bc, cFont::GetFont(fontOsd));

  sprintf(tmp, " %3i ", value);
  osd->DrawText(osd_width / 2 + 20, vert_start + line_nr * font_height, tmp, fc, bc, cFont::GetFont(fontOsd));
}

void cWhiteCalibration::DrawTexts(void)
{
  if (!osd) { return; }

  osd->DrawRectangle(0, 0, osd_width, osd_height, FCOLOR);

  DrawLine(1, tr("channel"), cur_channel);
  DrawLine(2, tr("red"),   SolarstormSetup.WhiteCalibration[cur_channel-1].r);
  DrawLine(3, tr("green"), SolarstormSetup.WhiteCalibration[cur_channel-1].g);
  DrawLine(4, tr("blue"),  SolarstormSetup.WhiteCalibration[cur_channel-1].b);

  DrawMark();
}

void cWhiteCalibration::Show(void)
{
  osd = cOsdProvider::NewOsd(0, 0);
  if (osd)
  {
    tArea Area = { 0, 0, osd_width, osd_height, 1 };
    osd->SetAreas(&Area, 1);
    DrawTexts();
    osd->Flush();
  }
}

bool cWhiteCalibration::SetupParse(const char *Name, const char *Value)
{
  if (!strcasecmp(Name, WHITECALIBRATION))
  {
    parameter = new char[strlen(Value)+1];
    strcpy(parameter, Value);
  }
  else return false;

  return true;
}

void cWhiteCalibration::Store(void)
{
  int sum = SolarstormSetup.ChannelsLeft + SolarstormSetup.ChannelsRight + SolarstormSetup.ChannelsTop + SolarstormSetup.ChannelsBottom;
  char *str = new char[12 * sum + 1];
  int pos = 0;

  for (int i=0; i<sum; i++)
  {
    pos += sprintf(str+pos, "%d ", SolarstormSetup.WhiteCalibration[i].r);
    pos += sprintf(str+pos, "%d ", SolarstormSetup.WhiteCalibration[i].g);
    pos += sprintf(str+pos, "%d ", SolarstormSetup.WhiteCalibration[i].b);
  }
  SolarstormPlugin->SetupStore(WHITECALIBRATION, str);

  delete [] str;
}

eOSState cWhiteCalibration::ProcessKey(eKeys Key)
{
  eOSState state = cOsdObject::ProcessKey(Key);

  if (state == osUnknown)
  {
    switch (Key & ~k_Repeat)
    {
      case kUp:     if (cur_line > 1)         { cur_line--; } break;
      case kDown:   if (cur_line < MAX_LINES) { cur_line++; } break;

      case kLeft:   Adjust( -1); break;
      case kRight:  Adjust(  1); break;
      case k1:      Adjust( -5); break;
      case k3:      Adjust(  5); break;
      case k4:      Adjust(-25); break;
      case k6:      Adjust( 25); break;
      case k7:      Adjust(-50); break;
      case k9:      Adjust( 50); break;

      case kOk:     Store(); return osEnd; break;
      // case kBack:   return osEnd; break;
      default:      return state; break;
    }
    DrawTexts();
    osd->Flush();
    state = osContinue;
  }
  return state;
}
